from typing import Dict, Tuple, Optional, Callable, Union, List, Set, Literal
import itertools
from collections import defaultdict
from copy import deepcopy
import os
import json

import pandas as pd
from loguru import logger

from syngen.ml.utils import (
    nan_labels_to_float,
    convert_date_to_timestamp,
    fetch_config
)
from syngen.ml.metrics import AccuracyTest, SampleAccuracyTest
from syngen.ml.data_loaders import DataLoader, DataFrameFetcher
from syngen.ml.metrics.utils import text_to_continuous
from syngen.ml.mlflow_tracker import MlflowTracker
from syngen.ml.utils import ProgressBarHandler


class Reporter:
    def __init__(
        self,
        table_name: str,
        paths: Dict[str, str],
        config: Dict[str, str],
        metadata: Dict,
        loader: Optional[Callable[[str], pd.DataFrame]] = None,
        type_of_process: Literal["train", "infer"] = "train"
    ):
        self.table_name = table_name
        self.paths = paths
        self.config = config
        self.metadata = deepcopy(metadata)
        self.loader = loader
        self.dataset = None
        self.columns_nan_labels = dict()
        self.na_values = dict()
        self.technical_columns = self._fetch_technical_columns()
        self.type_of_process = type_of_process

    def _fetch_technical_columns(self) -> Set[str]:
        """
        Check if there are technical columns in the dataframe and return them
        """
        path_to_flatten_metadata = self.paths.get("path_to_flatten_metadata")
        if os.path.exists(path_to_flatten_metadata):
            with open(path_to_flatten_metadata, "r") as f:
                flatten_metadata = json.load(f)
            if (flatten_metadata_of_table := flatten_metadata.get(self.table_name)) is not None:
                flattening_mapping = flatten_metadata_of_table.get("flattening_mapping")
                technical_columns = set(f"{col}_" for col in flattening_mapping.keys())
                return technical_columns
        return set()

    def _fetch_dataframe(self) -> pd.DataFrame:
        """
        Fetch the data using the callback function
        """
        data, schema = DataFrameFetcher(
            loader=self.loader,
            table_name=self.table_name
        ).load_data()
        logger.warning(
            f"The original data of the table - '{self.table_name}' "
            "has been fetched using the callback function. "
            "The data may have been modified since the beginning of the training process."
        )
        return data

    def _extract_report_data(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if self.loader:
            original = self._fetch_dataframe()
        else:
            original, schema = DataLoader(
                path=self.paths["input_data_path"],
                table_name=self.table_name,
                metadata=self.metadata,
                sensitive=True
            ).load_data()
        synthetic, schema = DataLoader(path=self.paths["path_to_merged_infer"]).load_data()
        synthetic = synthetic[
            [col for col in synthetic.columns if col not in self.technical_columns]
        ]
        original = original[
            [col for col in original.columns if col not in self.technical_columns]
        ]
        return original, synthetic

    def fetch_data_types(self):
        self.dataset = fetch_config(self.paths["dataset_pickle_path"])
        self.columns_nan_labels = self.dataset.nan_labels_dict
        types = (
            self.dataset.str_columns,
            self.dataset.date_columns,
            self.dataset.int_columns,
            self.dataset.float_columns,
            self.dataset.binary_columns,
            self.dataset.categorical_columns,
            self.dataset.long_text_columns,
            self.dataset.email_columns,
        )

        # eliminate keys columns from the report
        keys_columns = (
            set(self.dataset.pk_columns) |
            set(self.dataset.fk_columns) |
            set(self.dataset.uq_columns)
        )
        types = tuple(columns - keys_columns - self.technical_columns for columns in types)

        return types

    @staticmethod
    def convert_dates_to_timestamps(data: pd.Series, date_format: str, na_values: List[str]):
        """
        Convert the string values of dates to timestamps
        """
        return [
            convert_date_to_timestamp(d, date_format, na_values) for d in data
        ]

    def preprocess_data(self, original: pd.DataFrame, synthetic: pd.DataFrame):
        """
        Preprocess original and synthetic data.
        Return original data, synthetic data, float columns, integer columns, categorical columns
        without keys columns
        """
        types = self.fetch_data_types()
        exclude_columns = self.dataset.uuid_columns
        for column in self.dataset.cast_to_integer:
            original[column] = pd.to_numeric(
                original[column], errors="coerce", downcast="integer"
            )
            synthetic[column] = pd.to_numeric(
                synthetic[column], errors="coerce", downcast="integer"
            )
        for column in self.dataset.cast_to_float:
            original[column] = pd.to_numeric(
                original[column], errors="coerce", downcast="float"
            )
            synthetic[column] = pd.to_numeric(
                synthetic[column], errors="coerce", downcast="float"
            )
        original = nan_labels_to_float(
            original, self.columns_nan_labels, exclude_columns, process="report"
        )
        synthetic = nan_labels_to_float(
            synthetic, self.columns_nan_labels, exclude_columns, process="report"
        )
        (
            str_columns,
            date_columns,
            int_columns,
            float_columns,
            binary_columns,
            categorical_columns,
            long_text_columns,
            email_columns,
        ) = types

        original = original[[col for col in original.columns if col in set().union(*types)]]
        synthetic = synthetic[[col for col in synthetic.columns if col in set().union(*types)]]
        na_values = self.dataset.format.get("na_values", [])
        for date_col, date_format in self.dataset.date_mapping.items():
            original[date_col] = self.convert_dates_to_timestamps(
                original[date_col], date_format, na_values
            )
            synthetic[date_col] = self.convert_dates_to_timestamps(
                synthetic[date_col], date_format, na_values
            )

        int_columns = date_columns | int_columns
        text_columns = str_columns | long_text_columns | email_columns
        original = text_to_continuous(original, text_columns).drop(text_columns, axis=1)
        synthetic = text_to_continuous(synthetic, text_columns).drop(text_columns, axis=1)

        for col in [i + "_word_count" for i in text_columns]:
            if original[col].nunique() < 50:  # ToDo check if we need this
                categorical_columns = categorical_columns | {col}
            else:
                int_columns = int_columns | {col}
        int_columns = int_columns | {i + "_char_len" for i in text_columns}

        categorical_columns = categorical_columns | binary_columns

        for col in categorical_columns:
            original[col] = original[col].fillna("nan").astype(str)
            synthetic[col] = synthetic[col].fillna("nan").astype(str)
        return (
            original,
            synthetic,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        )


class Report:
    """
    Singleton metaclass for registration all needed reporters
    """

    _reporters: Dict[str, Union[Reporter, List]] = {}

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(Report, cls).__new__(cls)
        return cls.instance

    @classmethod
    def register_reporter(cls, table: str, reporter: Reporter):
        """
        Register all needed reporters
        """
        list_of_reporters = cls._reporters.get(table, [])
        list_of_reporters.append(reporter)
        cls._reporters[table] = list_of_reporters

    @classmethod
    def unregister_reporters(cls, table: str):
        """
        Unregister all reporters for a table
        """
        cls._reporters[table] = list()

    @classmethod
    def clear_report(cls):
        """
        Delete unnecessary reporters
        """
        cls._reporters.clear()

    @classmethod
    def _group_reporters(cls):
        """
        Group reporters by table name
        """
        list_of_reporters = itertools.chain.from_iterable(cls._reporters.values())
        sorted_reporters = sorted(list_of_reporters, key=lambda r: r.table_name)
        grouped_reporters = defaultdict(list)

        for reporter in sorted_reporters:
            grouped_reporters[reporter.table_name].append(reporter)

        return grouped_reporters

    @classmethod
    def generate_report(cls):
        """
        Generate all needed reports
        """
        grouped_reporters = cls._group_reporters()

        if grouped_reporters:
            logger.warning("The report(s) generation might be time-consuming")

        for table_name, reporters in grouped_reporters.items():
            cls._start_mlflow_run(table_name)
            delta = 0.25 / len(reporters)

            for reporter in reporters:
                cls._launch_reporter(reporter, delta)

            MlflowTracker().end_run()

    @staticmethod
    def _start_mlflow_run(table_name: str):
        MlflowTracker().start_run(
            run_name=f"{table_name}-REPORT",
            tags={"table_name": table_name, "process": "report"},
        )

    @classmethod
    def _launch_reporter(cls, reporter, delta: float):
        if reporter.__class__.report_type == "sample":
            # TODO: now the sample report isn't generated if the flatten metadata exists
            # This should be refactored in the future
            flatten_metadata_exists = os.path.exists(reporter.paths["path_to_flatten_metadata"])
            table_name = reporter.table_name
            if flatten_metadata_exists:
                logger.warning(
                    f"The sample report isn't available for a table - '{table_name}' "
                    "as it contains JSON column(s)."
                )
                return
            if reporter.loader is not None:
                logger.warning(
                    f"The sample report isn't available for a table - '{table_name}' "
                    "when a custom 'loader' is provided."
                )
                return

        cls._log_and_update_progress(
            delta,
            f"The calculation of {reporter.__class__.report_type} metrics for the table - "
            f"'{reporter.table_name}' has started"
        )

        reporter.report()

        if (
                reporter.__class__.report_type == "accuracy"
                and "accuracy" not in reporter.config["reports"]
                and "metrics_only" in reporter.config["reports"]
        ):
            message = (
                f"The metrics for the table - '{reporter.table_name}' have been evaluated"
            )
        else:
            message = (
                f"The {reporter.__class__.report_type} report of the table - "
                f"'{reporter.table_name}' has been generated"
            )
        cls._log_and_update_progress(delta, message)

    @staticmethod
    def _log_and_update_progress(delta: float, message: str):
        ProgressBarHandler().set_progress(delta=delta, message=message)
        logger.info(message)
        ProgressBarHandler().set_progress(
            progress=ProgressBarHandler().progress + delta,
            delta=delta,
            message=message
        )

    @property
    def reporters(self) -> Dict[str, Reporter]:
        return self._reporters


class AccuracyReporter(Reporter):
    """
    Reporter for running an accuracy test
    """

    report_type = "accuracy"

    def report(self):
        """
        Run the report
        """
        original, synthetic = self._extract_report_data()
        (
            original,
            synthetic,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        ) = self.preprocess_data(original, synthetic)
        accuracy_test = AccuracyTest(
            original=original,
            synthetic=synthetic,
            paths=self.paths,
            table_name=self.table_name,
            type_of_process=self.type_of_process,
            config=self.config
        )
        accuracy_test.report(
            cont_columns=list(float_columns | int_columns),
            categorical_columns=list(categorical_columns),
            date_columns=list(date_columns),
        )


class SampleAccuracyReporter(Reporter):
    """
    Reporter for running a sample test
    """

    report_type = "sample"

    def _extract_report_data(self):
        original, schema = DataLoader(path=self.paths["source_path"]).load_data()
        sampled, schema = DataLoader(
            path=self.paths["input_data_path"],
            table_name=self.table_name,
            metadata=self.metadata,
            sensitive=True
        ).load_data()
        return original, sampled

    def report(self):
        """
        Run the report
        """
        original, sampled = self._extract_report_data()
        (
            original,
            sampled,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        ) = self.preprocess_data(original, sampled)
        accuracy_test = SampleAccuracyTest(
            original,
            sampled,
            self.paths,
            self.table_name,
            self.config,
        )
        accuracy_test.report(
            cont_columns=list(float_columns | int_columns),
            categorical_columns=list(categorical_columns),
            date_columns=list(date_columns),
        )



import os
import sys
import re
from typing import List, Dict, Optional, Union, Set, Tuple, Literal
from dateutil import parser
from datetime import datetime, timedelta, date, timezone
import time
from pathlib import Path

import pandas as pd
import numpy as np
from slugify import slugify
import pickle as pkl
import uuid
from ulid import ULID
import random
from loguru import logger


MAX_ALLOWED_TIME_MS = 253402300800   # datetime(9999, 12, 31, 23, 59, 59, 999999).timestamp()
MIN_ALLOWED_TIME_MS = -62135510400   # datetime(1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)

# IANA timezone names - "2023-07-02T10:18:44.000000 America/New_York"
# Zulu time (UTC) represented by 'Z' - "2023-07-02T10:18:44Z"
# Numeric offsets (like +02:00, -05:00, etc.) - "2023-07-02T10:18:44 +02:00"
# Timezone abbreviations (like EST, PST, GMT, etc.) - "2023-07-02T10:18:44 EST"
TIMEZONE_REGEX = re.compile(r"""
        (?P<iana_name>
            [A-Za-z_]+
            (?:/[A-Za-z0-9_/-]+)+
        )|
        (?P<offset_zulu>
            Z
        )|
        (?P<offset_numeric>
        [+-]                         # Match a '+' or '-' for timezone offset
        (?:
        \d{2}:\d{2}$                 # Matches "+HH:MM" or "-HH:MM"
        |
        \d{4}$                       # Matches "+HHMM" or "-HHMM"
        )
        )|
        (?P<tz_abbr>
            \b
            (?!AM\b|PM\b|YT\b)
            \b
            (?:[A-Z]{2,5})
            \b
        )
    """, re.VERBOSE | re.ASCII)


class ProgressBarHandler:
    """
    Singleton class for handling the progress bar
    """
    _delta = None
    _progress = 0
    _message = None

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(ProgressBarHandler, cls).__new__(cls)
            cls.instance._progress = 0
            cls.instance._delta = None
            cls.instance._message = None
        return cls.instance

    @property
    def progress(self):
        """
        Get the current progress of the process
        """
        return self._progress

    @property
    def delta(self):
        """
        Get the delta of the progress
        inside which the progress should be changed
        """
        return self._delta

    @property
    def info(self):
        """
        Get the information about the current progress
        and the log message described the status of the process
        """
        return self._progress, self._message

    def set_progress(self, progress=None, delta=None, message=None):
        self._delta = delta if delta is not None else self._delta
        self._progress = self._progress if progress is None else progress
        self._progress = self._progress if self._progress <= 1.0 else 1.0
        self._message = message

    @classmethod
    def reset_instance(cls):
        cls.instance._progress = 0
        cls.instance._delta = None
        cls.instance._message = None


def is_format_first(date_format: str, format_type: str) -> bool:
    """
    Check if the date format starts with the specified string
    """
    return date_format.lower().startswith(f"%{format_type}")


def datetime_str_to_timestamp(dt: str, date_format: str) -> float:
    """
    Convert the datetime string to the timestamp
    """
    try:
        dt = parser.parse(
            dt,
            dayfirst=is_format_first(date_format, "d"),
            yearfirst=is_format_first(date_format, "y")
        ).replace(tzinfo=None)
        delta = dt - datetime(1970, 1, 1)
        return delta.total_seconds()
    except parser._parser.ParserError as e:
        year = re.match(r"\d+", e.args[0][5:]).group(0)
        if int(year) > 9999:
            return MAX_ALLOWED_TIME_MS
        elif int(year) < 1:
            return MIN_ALLOWED_TIME_MS


def datetime_to_timestamp(dt, date_format):
    """
    Convert datetime to timestamp
    """
    if pd.isnull(dt):
        return np.nan
    if isinstance(dt, str):
        return datetime_str_to_timestamp(dt, date_format)
    if isinstance(dt, datetime):
        return dt.timestamp()
    if isinstance(dt, date):
        # Convert date to datetime at midnight
        return datetime.combine(dt, datetime.min.time()).timestamp()


def timestamp_to_datetime(timestamp: int, delta=False):
    """
    Convert the timestamp to the datetime object or timedelta object
    """
    # Calculate the number of seconds in the UNIX epoch and the number of seconds left
    if pd.isnull(timestamp):
        return np.nan

    if timestamp >= MAX_ALLOWED_TIME_MS:
        return datetime(9999, 12, 31, 23, 59, 59, 999999)
    elif timestamp <= MIN_ALLOWED_TIME_MS:
        return datetime(1, 1, 1, 0, 0, tzinfo=timezone.utc)

    seconds_since_epoch = int(timestamp)
    remaining_seconds = timestamp - seconds_since_epoch

    # Calculate the datetime for the UNIX epoch (January 1, 1970)
    epoch_datetime = datetime(1970, 1, 1)

    # Calculate the timedelta for the number of seconds in the UNIX epoch
    delta_of_time = timedelta(seconds=seconds_since_epoch) + timedelta(seconds=remaining_seconds)

    if delta:
        return delta_of_time

    else:
        return epoch_datetime + delta_of_time


def convert_to_date(
    value: Union[int, float],
    date_format: str,
    to_datetime_conversion: bool = False
) -> Union[str, Union[datetime, timedelta]]:
    """
    Convert timestamp to date string values or values of date objects
    represented dates in a column
    """
    date_format = date_format if date_format else "%Y-%m-%d %H:%M:%S"
    dt = timestamp_to_datetime(int(value))
    if to_datetime_conversion:
        return dt
    return dt.strftime(date_format)


def generate_uuids(version: Union[int, str], size: int):
    ulid = ULID()
    generated_uuid_column = []
    for i in range(size):
        if version != "ulid":
            generated_uuid_column.append(
                str(uuid.UUID(int=random.getrandbits(128), version=int(version)))
            )
        else:
            generated_uuid_column.append(ulid.generate())
    return generated_uuid_column


def generate_uuid(size, dataset, uuid_columns, synthetic_infer):
    uuid_columns_types = dataset.uuid_columns_types
    for col in uuid_columns:
        synthetic_infer[col] = generate_uuids(uuid_columns_types[col], size)
    return synthetic_infer


def get_date_columns(df: pd.DataFrame, str_columns: List[str]):
    def date_finder(x, fuzzy=False):
        x_wo_na = x.dropna()
        count = 0
        for x in x_wo_na.values:
            try:
                date_for_check = datetime(8557, 7, 20)
                datetime_object = parser.parse(x, default=date_for_check)

                # Check if the parsed date contains only the time component.
                # If it does, then skip it.
                count += 1 if datetime_object.date() != date_for_check.date() else 0
            except (ValueError, OverflowError):
                continue

        if count == len(x_wo_na):
            return 1
        else:
            return np.nan

    date_columns = df[str_columns].apply(date_finder).dropna()

    if isinstance(date_columns, pd.DataFrame):
        names = date_columns.columns
    elif isinstance(date_columns, pd.Series):
        names = date_columns.index
    else:
        names = []

    return set(names)


def convert_date_to_timestamp(
    value,
    date_format: str,
    na_values: list
) -> float | None:
    """Helper to convert a single date value to a timestamp"""
    if value is None or value in na_values:
        return None
    result = datetime_to_timestamp(value, date_format)
    try:
        if np.isnan(result):
            return None
    except (TypeError, ValueError):
        pass
    return result


def fetch_timezone(date_string: str) -> Union[str, float]:
    """
    Attempts to find and extract a timezone string from a date string.

    Args:
        date_string: The input string that might contain a date and timezone.

    Returns:
        timezone_string:
        The matched timezone string (e.g., "America/New_York", "+05:30", "EST").
    """
    if not isinstance(date_string, str):
        return np.NaN

    match = TIMEZONE_REGEX.search(date_string)

    if match:
        if match.group("iana_name"):
            return match.group("iana_name")
        elif match.group("offset_zulu"):
            return match.group("offset_zulu")
        elif match.group("offset_numeric"):
            return match.group("offset_numeric")
        elif match.group("tz_abbr"):
            return match.group("tz_abbr")

    return np.NaN


def get_nan_labels(df: pd.DataFrame, excluded_columns: Set[str]) -> Dict:
    """
    Get labels that represent nan values in float/int columns
    """
    columns_nan_labels = {}
    object_columns = df.select_dtypes(include=[pd.StringDtype(), "object", "string"]).columns
    columns = set(object_columns) - excluded_columns
    for column in columns:
        if df[column].isna().sum() > 0:
            continue
        str_values = []
        float_val = None
        for val in df[column].unique():
            try:
                float_val = float(val)
            except (TypeError, ValueError):
                str_values.append(val)
        if (float_val is not None) and (not np.isnan(float_val)) and len(str_values) == 1:
            nan_label = str_values[0]
            columns_nan_labels[column] = nan_label

    return columns_nan_labels


def nan_labels_to_float(
    df: pd.DataFrame,
    columns_nan_labels: dict,
    exclude_columns: set = set(),
    process="training"
) -> pd.DataFrame:
    """
    Replace str nan labels in float/int columns with actual np.NaN
    and casting the column to float type.
    """
    df_with_nan = df.copy()
    for column, label in columns_nan_labels.items():
        if column not in exclude_columns:
            df_with_nan[column].replace(label, np.NaN, inplace=True)
            df_with_nan[column] = df_with_nan[column].astype(float)
            if process == "training":
                logger.info(
                    f"Column '{column}' contains unique "
                    f"non-numeric value: '{label}'. "
                    "It will be treated as null label "
                    "and replaced with nulls "
                    "during the training process."
                )

    return df_with_nan


def fillnan(df, str_columns, float_columns, categorical_columns):
    for c in str_columns | categorical_columns:
        df[c] = df[c].fillna("NaN")

    return df


def slugify_attribute(**kwargs):
    """
    Slugify the value of the attribute of the instance
    and set it to the new attribute
    """

    def wrapper(function):
        def inner_wrapper(*args):
            object_, *other = args
            for attribute, new_attribute in kwargs.items():
                fetched_attribute = object_.__getattribute__(attribute)
                value_of_new_attribute = slugify(fetched_attribute)
                object_.__setattr__(new_attribute, value_of_new_attribute)
            return function(*args)

        return inner_wrapper

    return wrapper


def slugify_parameters(exclude_params=(), turn_on=True):
    """
    Slugify the values of parameters, excluding specified parameters
    """

    def wrapper(function):
        def inner_wrapper(**kwargs):
            updated_kwargs = {}
            for key, value in kwargs.items():
                if key in exclude_params:
                    updated_kwargs[key] = value
                else:
                    updated_kwargs[key] = slugify(value, lowercase=turn_on)
            return function(**updated_kwargs)

        return inner_wrapper

    return wrapper


def inverse_dict(dictionary: Dict) -> Dict:
    """
    Swap keys and values in the dictionary
    """
    return dict(zip(dictionary.values(), dictionary.keys()))


def clean_up_metadata(metadata: Dict):
    """
    Clean up the metadata,
    remove the sensitive information (credentials to the remote storage) from the metadata
    """
    forbidden_keys = ["data", "credentials", "fernet_key"]
    for key, value in list(metadata.items()):
        if key in forbidden_keys:
            del metadata[key]
        elif isinstance(value, dict):
            clean_up_metadata(value)
        else:
            continue
    return metadata


def trim_string(col):
    if isinstance(col.dtype, str):
        return col.str.slice(stop=10 * 1024)
    else:
        return col


def check_if_features_assigned(dataset_pickle_path: str):
    """
    Check if features are assigned in the dataset
    """
    features = fetch_config(dataset_pickle_path).features
    if len(features) == 0:
        logger.info("No features to train VAE on")
        return False
    return True


def fetch_config(config_pickle_path: str):
    """
    Fetch the configuration from the disk
    """
    with open(config_pickle_path, "rb") as f:
        return pkl.load(f)


def save_config(config_pickle_path: str, config):
    """
    Save the configuration to the disk
    """
    with open(config_pickle_path, "wb") as f:
        pkl.dump(config, f)


def fetch_unique_root(table_name: Optional[str], metadata_path: Optional[str]):
    """
    Construct the unique constant substring for use in the name of the experiment and log file
    """
    unique_name = str()
    if metadata_path:
        unique_name = os.path.basename(metadata_path)
    elif table_name:
        unique_name = table_name
    return slugify(unique_name)


def get_log_path(table_name: Optional[str], metadata_path: Optional[str], type_of_process: str):
    """
    Get the log path for storing the logs of main processes
    """
    logs_dir_name = "model_artifacts/system_store/logs"
    unique_name = fetch_unique_root(table_name, metadata_path)
    unique_name = f"{unique_name}_{datetime.now().strftime('%Y_%m_%d_%H_%M_%S_%f')}"
    file_name_without_extension = f"logs_{type_of_process}_{unique_name}"
    file_path = os.path.join(
        logs_dir_name, f"{slugify(file_name_without_extension)}.log"
    )
    return file_path


def fetch_log_message(message):
    """
    Fetch the log message
    """
    record = message.record
    log_message = (f'{record["time"]} | {record["level"]} | '
                   f'{record["file"]}:{record["function"]}:{record["line"]} - {record["message"]}')
    return log_message


def fetch_env_variables(options: Dict) -> Dict:
    """
    Fetches the values of environment variables specified in options
    """
    errors = []
    options = options.copy()
    for option, value in options.items():
        if value is not None:
            fetched_value = os.getenv(value)
            if fetched_value is None:
                error_message = (
                    f"The value of the environment variable '{value}' wasn't fetched. "
                    "Please, check whether it is set correctly."
                )
                errors.append(error_message)
            options[option] = fetched_value
    if errors:
        errors = " ".join(errors)
        raise ValueError(errors)
    return options


def file_sink(message):
    """
    Save logs to the log file
    """
    path_to_logs = os.getenv("SUCCESS_LOG_FILE")
    if path_to_logs is not None:
        with open(path_to_logs, "a") as log_file:
            log_message = fetch_log_message(message)
            log_file.write(log_message + "\n")


def console_sink(record):
    """
    Redirect logs to the console
    """
    sys.stderr.write(record)


def setup_logger():
    """
    Set up loggers with the specified level
    """
    logger.remove()
    logger.add(console_sink, colorize=True, level=os.getenv("LOGURU_LEVEL"))
    logger.add(file_sink, level=os.getenv("LOGURU_LEVEL"))


def setup_log_process(
    type_of_process: str,
    log_level: str,
    table_name: Optional[str],
    metadata_path: Optional[str]
):
    """
    Set up the logging process with the specified level
    """
    os.environ["LOGURU_LEVEL"] = log_level
    os.makedirs("model_artifacts/system_store/logs", exist_ok=True)
    os.environ["SUCCESS_LOG_FILE"] = get_log_path(
        table_name=table_name,
        metadata_path=metadata_path,
        type_of_process=type_of_process
    )
    setup_logger()


def get_initial_table_name(table_name) -> str:
    """"
    Get the initial table name without the suffix "_pk" or "_fk"
    """
    return re.sub(r"_pk$|_fk$", "", table_name)


def timing(func=None, log_level="TRACE"):
    """
    Decorator that logs the execution time of the function
    """
    def decorator(inner_func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = inner_func(*args, **kwargs)
            end_time = time.time()
            elapsed_time = end_time - start_time
            log_method = getattr(logger, log_level.lower(), logger.trace)
            log_method(f"Function '{inner_func.__name__}' executed in {elapsed_time:.2f} seconds.")
            return result

        return wrapper

    if func is not None and callable(func):
        # Used as @timing
        return decorator(func)
    else:
        # Used as @timing(...)
        return decorator


def get_reports(
    value: Union[List[str], Tuple[str], str],
    report_types,
    type_of_process: Literal["train", "infer"]
) -> List[str]:
    """
    Validate the values provided by the parameter 'value',
    convert it, and get the appropriate list of reports
    """
    list_of_report_types = (
        report_types.train_report_types
        if type_of_process == "train"
        else report_types.infer_report_types
    )
    full_list_of_report_types = (
        report_types.full_list_of_train_report_types
        if type_of_process == "train"
        else report_types.full_list_of_infer_report_types
    )
    input_values = set(value) if isinstance(value, (list, tuple)) else {value}
    valid_values: List = ["none", "all"]
    valid_values.extend(list_of_report_types)

    if not input_values.issubset(set(valid_values)):
        list_of_valid_values = ", ".join([f"'{value}'" for value in valid_values])
        raise ValueError(
            "Invalid input: Acceptable values for the parameter 'reports' are "
            f"{list_of_valid_values}."
        )
    if "none" in input_values and "all" in input_values:
        raise ValueError(
            "Invalid input: The 'reports' parameter cannot be set to both 'none' and 'all'. "
            "Please provide only one of these options."
        )

    if "none" in input_values or "all" in input_values:
        if len(input_values) > 1:
            raise ValueError(
                "Invalid input: When 'reports' option is set to 'none' or 'all', "
                "no other values should be provided."
            )
        if list(input_values)[0] == "all":
            return full_list_of_report_types
        if list(input_values)[0] == "none":
            return list()

    if {"accuracy", "metrics_only"}.issubset(input_values):
        return [v for v in input_values if v != "metrics_only"]

    return list(input_values)


class ValidationError(Exception):
    """
    The exception class for handling validation errors
    """
    def __init__(
        self,
        message: str
    ):
        super().__init__(message)
        self.message = message


def get_source_path_extension(
    table_name: Optional[str] = None,
    metadata: Dict = dict(),
    path: Optional[str] = None
) -> str:
    """
    Get the extension of the source
    """
    if table_name:
        source = (
            metadata.get(table_name, {}).get("train_settings", {}).get("source")
        )
    else:
        source = path
    return Path(source).suffix if source is not None else ".csv"



from typing import Dict, Tuple
from dataclasses import dataclass
from datetime import datetime, date

import pandas as pd
import numpy as np
from loguru import logger


@dataclass
class Convertor:
    """
    Abstract class for converting fetched schema in Avro, Parquet or Delta formats
    """
    schema: Dict
    df: pd.DataFrame
    excluded_dtypes: Tuple = (str, bytes, datetime, date)

    def _check_dtype_or_nan(self, dtypes: Tuple):
        """
        Check if the value is of the specified data types or 'np.NaN'
        """
        return (
            lambda x: isinstance(x, dtypes)
            or (not isinstance(x, self.excluded_dtypes) and np.isnan(x))
        )

    def _update_data_types(self, schema: Dict, df: pd.DataFrame):
        """
        Update data types related to the fetched schema
        """
        type_map = {
            "binary": "string",
            "string": "string",
            "double": "float64",
            "float": "float64",
            "decimal": "float64"
        }

        for column, data_type in schema.get("fields", {}).items():
            if column not in df.columns:
                continue

            if data_type == "null":
                if df[column].isnull().all():
                    continue
                else:
                    raise ValueError(
                        f"It seems that the data type - '{data_type}' "
                        f"isn't correct for the column - '{column}' as it's not empty"
                    )

            if data_type in type_map:
                df[column] = df[column].astype(type_map[data_type])
            elif data_type in ["int", "bool"]:
                if df[column].isnull().any():
                    df[column] = df[column].astype("float64")
                else:
                    df[column] = df[column].astype("int64")

        if not schema.get("fields"):
            for column in df.columns:
                if df[column].apply(lambda x: isinstance(x, int)).all():
                    df[column] = df[column].astype(int)
                elif df[column].apply(self._check_dtype_or_nan(dtypes=(int, float))).all():
                    df[column] = df[column].astype(float)
                elif df[column].apply(self._check_dtype_or_nan(dtypes=(str, bytes))).all():
                    df[column] = df[column].astype(pd.StringDtype())
        return df

    @staticmethod
    def _set_none_values_to_nan(df: pd.DataFrame):
        """
        Set 'None' values contained in columns as 'np.NaN'
        """
        df_object_subset = df.select_dtypes(["object"])
        for column in df_object_subset:
            df[column] = [
                np.NaN
                if i is None
                else i
                for i in df[column]
            ]
        return df

    def _cast_values_to_string(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Cast the values contained in columns with the data type 'object'
        to 'string'
        """
        df_object_subset = df.select_dtypes(["object"])
        for column in df_object_subset:
            df[column] = [
                i
                if (not isinstance(i, self.excluded_dtypes) and np.isnan(i))
                or isinstance(i, self.excluded_dtypes)
                else str(i)
                for i in df[column]
            ]
        return df

    def _preprocess_df(self, schema: Dict, df: pd.DataFrame) -> pd.DataFrame:
        """
        Preprocess data frame, update data types of columns
        """
        if not df.empty:
            try:
                df = self._set_none_values_to_nan(df)
                df = self._cast_values_to_string(df)
                df = self._update_data_types(schema, df)
                return df
            except Exception as e:
                logger.error(e)
                raise e
        else:
            return df


class CSVConvertor(Convertor):
    """
    Class for supporting custom schema for csv files
    """
    schema: Dict = {"fields": {}, "format": "CSV"}

    def __init__(self, df):
        schema = {"fields": {}, "format": "CSV"}
        super().__init__(schema, df)
        self.preprocessed_df = self._preprocess_df(schema, df)


class AvroConvertor(Convertor):
    """
    Class for converting fetched avro schema
    """

    def __init__(self, schema, df):
        super().__init__(schema, df)
        self.converted_schema = self._convert_schema(schema)
        self.preprocessed_df = self._preprocess_df(self.converted_schema, df)

    @staticmethod
    def _convert_schema(schema) -> Dict:
        """
        Convert the schema of Avro file to unified format, preprocess dataframe
        """
        converted_schema = dict()
        converted_schema["fields"] = dict()
        schema = schema if schema else dict()
        for column, data_type in schema.items():
            fields = converted_schema["fields"]
            if "int" in data_type or "long" in data_type or "boolean" in data_type:
                fields[column] = "int"
            elif "float" in data_type or "double" in data_type:
                fields[column] = "float"
            elif "string" in data_type:
                fields[column] = "string"
            elif "bytes" in data_type:
                fields[column] = "string"
            elif "null" in data_type:
                fields[column] = "null"
            else:
                message = (
                    f"It seems that the column - '{column}' has unsupported data type - "
                    f"'{data_type}'"
                )
                logger.error(message)
                raise ValueError(message)
        converted_schema["format"] = "Avro"
        return converted_schema



import os
import argparse
from syngen.sdk import Syngen  # noqa: F401

base_dir = os.path.dirname(__file__)
version_file = os.path.join(base_dir, "VERSION")

with open(version_file) as f:
    __version__ = f.read().strip()


def main():
    parser = argparse.ArgumentParser(prog="syngen")
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__
    )
    args = parser.parse_args()
    return args



from typing import Optional, List, Union, Tuple, Callable

import click
from loguru import logger
import pandas as pd

from syngen.ml.worker import Worker
from syngen.ml.utils import (
    setup_log_process,
    get_reports,
    fetch_env_variables
)
from syngen.ml.validation_schema import ReportTypes


def validate_required_parameters(
    metadata_path: Optional[str] = None,
    table_name: Optional[str] = None
):
    if not metadata_path and not table_name:
        raise AttributeError(
            "It seems that the information about 'metadata_path' or 'table_name' is absent. "
            "Please provide either the information about 'metadata_path' or the information "
            "about 'table_name'."
        )
    if metadata_path and table_name:
        logger.warning(
            "The information about 'metadata_path' was provided. "
            "In this case the information about 'table_name' will be ignored."
        )


def launch_infer(
    metadata_path: Optional[str] = None,
    size: Optional[int] = 100,
    table_name: Optional[str] = None,
    run_parallel: bool = False,
    batch_size: Optional[int] = None,
    reports: Union[List[str], Tuple[str], str] = "none",
    random_seed: Optional[int] = None,
    log_level: str = "INFO",
    fernet_key: Optional[str] = None,
    loader: Optional[Callable[[str], pd.DataFrame]] = None
):
    setup_log_process(
        type_of_process="infer",
        log_level=log_level,
        table_name=table_name,
        metadata_path=metadata_path
    )

    encryption_settings = fetch_env_variables({"fernet_key": fernet_key})

    worker = Worker(
        table_name=table_name,
        metadata_path=metadata_path,
        settings={
            "size": size,
            "run_parallel": run_parallel,
            "batch_size": batch_size,
            "reports": get_reports(
                value=reports,
                report_types=ReportTypes(),
                type_of_process="infer"
            ),
            "random_seed": random_seed
        },
        log_level=log_level,
        type_of_process="infer",
        encryption_settings=encryption_settings,
        loader=loader
    )

    logger.info(
        "The inference process will be executed according to the information mentioned "
        "in 'infer_settings' in the metadata file. If appropriate information is absent "
        "from the metadata file, then the values of parameters sent through CLI will be used. "
        "Otherwise, the values of parameters will be defaulted."
    )

    worker.launch_infer()


@click.command()
@click.option(
    "--metadata_path",
    type=str,
    default=None,
    help="Path to the metadata file"
)
@click.option(
    "--size",
    default=100,
    type=click.IntRange(1),
    help="Desired number of rows to generate. If absent, it's defaulted to 100",
)
@click.option(
    "--table_name",
    default=None,
    type=str,
    help="Name of the table, same as in training",
)
@click.option(
    "--run_parallel",
    default=False,
    type=click.BOOL,
    help="The flag which set whether to use multiprocessing (feasible for tables > 5000 rows)."
    "If absent, it's defaulted to False",
)
@click.option(
    "--batch_size",
    default=None,
    type=click.IntRange(1),
    help="If specified, the generation is split into batches. This can save the RAM",
)
@click.option(
    "--random_seed",
    default=None,
    type=click.IntRange(0),
    help="Set any int in case you want reproducible results. To reproduce generated data again, "
    "use the same int in this command.",
)
@click.option(
    "--reports",
    default=("none",),
    type=click.UNPROCESSED,
    multiple=True,
    help="Controls the generation of quality reports. "
    "Might require significant time for big generated tables (>10000 rows). "
    "If set to 'accuracy', generates an accuracy report. "
    "If set to 'metrics_only', outputs the metrics information "
    "only to standard output without generation of a report. "
    "If set to 'all', generates an accuracy report. "
    "If it's absent or set to 'none', no reports are generated.",
)
@click.option(
    "--log_level",
    default="INFO",
    type=click.Choice(["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    help="Set the logging level which will be used in the process. "
         "If absent, it's defaulted to 'INFO'",
)
@click.option(
    "--fernet_key",
    default=None,
    type=str,
    help="The name of the environment variable that kept the value of the Fernet key "
         "to decrypt the sensitive data stored on the disk",
)
def cli_launch_infer(
    metadata_path: Optional[str],
    size: Optional[int],
    table_name: Optional[str],
    run_parallel: bool,
    batch_size: Optional[int],
    reports: List[str],
    random_seed: Optional[int],
    log_level: str,
    fernet_key: Optional[str]
):
    """
    Launch the work of infer process
    Parameters
    ----------
    metadata_path
    size
    table_name
    run_parallel
    batch_size
    reports
    random_seed
    log_level
    fernet_key
    -------
    """
    validate_required_parameters(metadata_path=metadata_path, table_name=table_name)

    launch_infer(
        metadata_path=metadata_path,
        size=size,
        table_name=table_name,
        run_parallel=run_parallel,
        batch_size=batch_size,
        reports=reports,
        random_seed=random_seed,
        log_level=log_level,
        fernet_key=fernet_key
    )


if __name__ == "__main__":
    cli_launch_infer()



import copy
import re
from math import isnan
from collections.abc import Iterable
import json
from json import JSONDecodeError


def check_if_numbers_are_consecutive(list_):
    """
    Returns True if numbers in the list are consecutive

    :param list_: list of integers
    :return: Boolean
    """
    return all(
        True if second - first == 1 else False
        for first, second in zip(list_[:-1], list_[1:])
    )


def _construct_key(previous_key, separator, new_key, replace_separators=None):
    """
    Returns the new_key if no previous key exists, otherwise concatenates
    previous key, separator, and new_key
    :param previous_key:
    :param separator:
    :param new_key:
    :param str replace_separators: Replace separators within keys
    :return: a string if previous_key exists and simply passes through the
    new_key otherwise
    """
    if replace_separators is not None:
        new_key = str(new_key).replace(separator, replace_separators)
    if previous_key:
        return u"{}{}{}".format(previous_key, separator, new_key)
    else:
        return new_key


def flatten(
    nested_dict,
    separator="_",
    root_keys_to_ignore=None,
    replace_separators=None
):
    """
    Flattens a dictionary with nested structure to a dictionary with no
    hierarchy
    Consider ignoring keys that you are not interested in to prevent
    unnecessary processing
    This is specially true for very deep objects

    :param nested_dict: dictionary we want to flatten
    :param separator: string to separate dictionary keys by
    :param root_keys_to_ignore: set of root keys to ignore from flattening
    :param str replace_separators: Replace separators within keys
    :return: flattened dictionary
    """
    assert isinstance(nested_dict, dict), "flatten requires a dictionary input"
    assert isinstance(separator, str), "separator must be string"

    if root_keys_to_ignore is None:
        root_keys_to_ignore = set()

    if len(nested_dict) == 0:
        return {}

    # This global dictionary stores the flattened keys and values and is
    # ultimately returned
    flattened_dict = dict()

    def _flatten(object_, key):
        """
        For dict, list and set objects_ calls itself on the elements and for
        other types assigns the object_ to
        the corresponding key in the global flattened_dict
        :param object_: object to flatten
        :param key: carries the concatenated key for the object_
        :return: None
        """
        # Empty object can't be iterated, take as is
        if not object_:
            flattened_dict[key] = object_
        # These object types support iteration
        elif isinstance(object_, dict):
            for object_key in object_:
                if not (not key and object_key in root_keys_to_ignore):
                    _flatten(
                        object_[object_key],
                        _construct_key(
                            key,
                            separator,
                            object_key,
                            replace_separators=replace_separators))
        elif isinstance(object_, (list, set, tuple)):
            for index, item in enumerate(object_):
                _flatten(
                    item,
                    _construct_key(
                        key,
                        separator,
                        index,
                        replace_separators=replace_separators))
        # Anything left take as is
        else:
            flattened_dict[key] = object_

    _flatten(nested_dict, None)
    return flattened_dict


def flatten_preserve_lists(nested_dict, separator="_",
                           root_keys_to_ignore=None,
                           max_list_index=3, max_depth=3,
                           replace_separators=None):
    """
    Flattens a dictionary with nested structure to a dictionary with no
    hierarchy
    Consider ignoring keys that you are not interested in to prevent
    unnecessary processing
    This is specially true for very deep objects
    This preserves list structure, and
    you can specify max_list_index and max_depth to limit processing

    Child elements with only one value inside
    will be unwrapped and become parent's value.

    :param nested_dict: dictionary we want to flatten
    :param separator: string to separate dictionary keys by
    :param root_keys_to_ignore: set of root keys to ignore from flattening
    :param max_list_index: maximum list index to process
    :param max_depth: maximum nesting depth to process
    :param str replace_separators: Replace separators within keys
    :return: flattened dictionary
    """

    assert isinstance(nested_dict, dict), "flatten requires a dictionary input"
    assert isinstance(separator, str), "separator must be a string"

    if root_keys_to_ignore is None:
        root_keys_to_ignore = set()

    # This global dictionary stores the flattened keys and values and is
    # ultimately returned
    flattened_dict = dict()

    def _flatten(object_, key):
        """
        For dict, list and set objects_ calls itself on the elements and for
        other types assigns the object_ to
        the corresponding key in the global flattened_dict
        :param object_: object to flatten
        :param key: carries the concatenated key for the object_
        :return: None
        """

        # Empty object can't be iterated, take as is
        if not object_:
            flattened_dict[key] = object_

        # These object types support iteration
        # dict always go into columns
        elif isinstance(object_, dict):
            first_key = list(object_.keys())[0]
            # if only 1 child value, and child value not a dict or list
            # flatten immediately
            is_iter = isinstance(object_[first_key], Iterable)
            if len(object_) == 1 and not is_iter:
                flattened_dict[key] = object_[first_key]
            else:
                for object_key in object_:
                    if not (not key and object_key in root_keys_to_ignore):
                        _flatten(
                            object_[object_key],
                            _construct_key(
                                key,
                                separator,
                                object_key,
                                replace_separators=replace_separators))

        elif isinstance(object_, (list, set, tuple)):
            for index, item in enumerate(object_):
                key = _construct_key(key, separator, index,
                                     replace_separators=replace_separators)
                _flatten(item, key)

        else:
            flattened_dict[key] = object_

    def _flatten_low_entropy(object_, key, cur_depth, max_depth_inner):
        """
        For dict, list and set objects_ calls itself on the elements and for
        other types assigns the object_ to
        the corresponding key in the global flattened_dict

        :param object_: object to flatten
        :param key: carries the concatenated key for the object_
        :return: None
        """
        cur_depth = cur_depth + 1  # increase current_depth
        debug = 0

        # write latest child as value if max_depth exceeded
        if cur_depth > max_depth_inner:
            global_max_record = int(max(list(
                list_prebuilt_flattened_dict.keys())))
            for d in list_prebuilt_flattened_dict[str(global_max_record)]:
                d[key] = object_

        else:
            # Empty object can't be iterated, take as is
            if not object_:
                global_max_record = int(max(list(
                    list_prebuilt_flattened_dict.keys())))
                for d in list_prebuilt_flattened_dict[str(global_max_record)]:
                    d[key] = object_

            # These object types support iteration
            # dict always go into columns
            elif isinstance(object_, dict):
                first_key = list(object_.keys())[0]
                # if only 1 child value, and child value
                # not a dict or list, flatten immediately
                if len(object_) == 1 \
                        and not (isinstance(object_[first_key], dict)
                                 or isinstance(object_[first_key], list)):
                    global_max_record = int(max(list(
                        list_prebuilt_flattened_dict.keys())))

                    for d in list_prebuilt_flattened_dict[
                        str(global_max_record)
                    ]:
                        d[key] = object_[first_key]

                else:
                    for object_key, val in \
                            sorted(object_.items(),
                                   key=lambda x:
                                   (str(type(x[1])), len(str(x[1]))),
                                   reverse=False):
                        if not (not key and object_key in root_keys_to_ignore):
                            _flatten_low_entropy(
                                object_[object_key],
                                _construct_key(
                                    key,
                                    separator,
                                    object_key,
                                    replace_separators=replace_separators),
                                cur_depth,
                                max_depth_inner)

            # lists could go into rows, like in a relational database
            elif isinstance(object_, list) or isinstance(object_, set):
                if debug:
                    print("\nparent key of list:",
                          key, "| length: ",
                          str(len(object_)))

                # need to remember global list state when we entered
                # this recursion
                global_max_record_start = int(max(list(
                    list_prebuilt_flattened_dict.keys())))
                entry = copy.deepcopy(list_prebuilt_flattened_dict[
                                          str(global_max_record_start)
                                      ])

                for index, item in enumerate(object_):

                    if debug:
                        print("  list key:", key,
                              " index: " + str(index), "vals: ", item)

                    sub = -1
                    if isinstance(item, dict):
                        first_value = list(item.values())[0]
                        if isinstance(first_value, float):
                            sub = first_value

                    if not isnan(sub) and index < max_list_index:
                        # start from second element, 1st element is like column
                        if index > 0:
                            global_max_record = int(max(list(
                                list_prebuilt_flattened_dict.keys())))

                            list_prebuilt_flattened_dict[
                                str(global_max_record + 1)
                            ] = copy.deepcopy(entry)

                        _flatten_low_entropy(item, key, cur_depth,
                                             max_depth_inner)
                    else:
                        pass

                list_prebuilt_flattened_dict['0'] = \
                    [subel for k, v in
                     sorted(list_prebuilt_flattened_dict.items())
                     for idx, subel in enumerate(v)]

                for key in list(sorted(list_prebuilt_flattened_dict.keys())):
                    if key != '0':
                        del list_prebuilt_flattened_dict[key]
                if debug:
                    print("collapsed global list")

            # Anything left take as is, assuming you hit the end of the line.
            else:
                # in this case, there may be
                # a list of prebuilt_flattened_dict by now
                # so need to update them all.
                global_max_record = int(max(list(
                    list_prebuilt_flattened_dict.keys())))

                for d in list_prebuilt_flattened_dict[str(global_max_record)]:
                    d[key] = object_

                    # decrease depth counter
        cur_depth -= 1

    _flatten(nested_dict, None)

    # get unique column names, without the integers
    # TODO: potential issue: what if column names have digits naturally?
    reskeys = list(flattened_dict.keys())
    unique_integers = list(set([separator + char for key
                                in reskeys for char in key if char.isdigit()]))
    regex = '|'.join(unique_integers)
    regex += "|" + regex.replace(".", "")
    unique_columns = list(set([re.sub("(" + regex + ")", "", key)
                               for key in reskeys]))

    # create global dict, now with unique column names
    prebuilt_flattened_dict = {column: None for column in unique_columns}

    # initialize global record list
    list_prebuilt_flattened_dict = {'0': [prebuilt_flattened_dict]}

    _flatten_low_entropy(nested_dict, None, cur_depth=0,
                         max_depth_inner=max_depth)

    return list_prebuilt_flattened_dict['0']


def _unflatten_asserts(flat_dict, separator):
    assert isinstance(flat_dict, dict), "un_flatten requires dictionary input"
    assert isinstance(separator, str), "separator must be string"
    assert all((not value or not isinstance(value, Iterable) or
                isinstance(value, str)
                for value in flat_dict.values())), "provided dict is not flat"


def unflatten(flat_dict, separator='_'):
    """
    Creates a hierarchical dictionary from a flattened dictionary
    Assumes no lists are present
    :param flat_dict: a dictionary with no hierarchy
    :param separator: a string that separates keys
    :return: a dictionary with hierarchy
    """
    _unflatten_asserts(flat_dict, separator)

    # This global dictionary is mutated and returned
    unflattened_dict = dict()

    def _unflatten(dic, keys, value):
        cur = dic
        for key in keys[:-1]:
            # ensure intermediate nodes are dicts; replace non-dict values so setdefault-like
            # behavior doesn't attempt to call methods on non-dict objects (e.g. float).
            if key not in cur or not isinstance(cur[key], dict):
                cur[key] = {}
            cur = cur[key]

        cur[keys[-1]] = value

    list_keys = sorted(flat_dict.keys())
    for i, item in enumerate(list_keys):
        if i != len(list_keys) - 1:
            split_key = item.split(separator)
            next_split_key = list_keys[i + 1].split(separator)
            if not split_key == next_split_key[:-1]:
                _unflatten(unflattened_dict, item.split(separator),
                           flat_dict[item])
            else:
                pass  # if key contained in next key, json will be invalid.
        else:
            #  last element
            _unflatten(unflattened_dict, item.split(separator),
                       flat_dict[item])
    return unflattened_dict


def unflatten_list(flat_dict, separator='_'):
    """
    Unflattens a dictionary, first assuming no lists exist and then tries to
    identify lists and replaces them
    This is probably not very efficient and has not been tested extensively
    Feel free to add test cases or rewrite the logic
    Issues that stand out to me:
    - Sorting all the keys in the dictionary, which specially for the root
    dictionary can be a lot of keys
    - Checking that numbers are consecutive is O(N) in number of keys

    :param flat_dict: dictionary with no hierarchy
    :param separator: a string that separates keys
    :return: a dictionary with hierarchy
    """
    _unflatten_asserts(flat_dict, separator)

    # First unflatten the dictionary assuming no lists exist
    unflattened_dict = unflatten(flat_dict, separator)

    def _convert_dict_to_list(object_, parent_object, parent_object_key):
        if isinstance(object_, dict):
            for key in object_:
                if isinstance(object_[key], dict):
                    _convert_dict_to_list(object_[key], object_, key)
            try:
                keys = [int(key) for key in object_]
                keys.sort()
            except (ValueError, TypeError):
                keys = []
            keys_len = len(keys)

            if (keys_len > 0 and sum(keys) ==
                    int(((keys_len - 1) * keys_len) / 2) and keys[0] == 0 and
                    keys[-1] == keys_len - 1 and
                    check_if_numbers_are_consecutive(keys)):

                # The dictionary looks like a list so we're going to replace it
                parent_object[parent_object_key] = []
                for key_index, key in enumerate(keys):
                    parent_object[parent_object_key].append(object_[str(key)])
                    _convert_dict_to_list(parent_object[parent_object_key][-1],
                                          parent_object[parent_object_key],
                                          key_index)

    _convert_dict_to_list(unflattened_dict, None, None)
    return unflattened_dict


def safe_flatten(val):
    """
    Safely flatten a JSON string into a flat dictionary.
    If the input won't be flattened, return the original value.
    """
    if not isinstance(val, (str, bytes, bytearray)):
        return {
            "original_data": val,
            "flattened_data": {}
        }

    try:
        parsed = json.loads(val)
        if isinstance(parsed, dict):
            return {
                "flattened_data": flatten(parsed, "."),
                "original_data": None
            }
    except JSONDecodeError:
        pass

    return {
        "original_data": val,
        "flattened_data": {}
    }



import os
from collections import Counter
from typing import List, Tuple, Dict, Any, Optional, Callable, Union, Literal
import json
from json import JSONDecodeError
from slugify import slugify

from loguru import logger
import pandas as pd
import numpy as np
from syngen.ml.flatten_json import unflatten_list, safe_flatten

from syngen.ml.data_loaders import DataLoader, DataFrameFetcher
from syngen.ml.utils import fetch_unique_root, fetch_config, get_source_path_extension
from syngen.ml.context import get_context


class Processor:
    """
    Base class for data preprocessing and postprocessing

    Attributes:
        metadata (Dict): Processing metadata
        metadata_path (Optional[str]): Path to metadata file
        table_name (Optional[str]): Name of the table being processed
        loader (Optional[Callable]): Custom data loader function
    """
    def __init__(
        self,
        metadata: Dict,
        metadata_path: Optional[str],
        table_name: str,
        loader: Optional[Callable[[str], pd.DataFrame]] = None
    ):
        self.metadata_path = metadata_path
        self.metadata = metadata
        self.table_name = table_name
        self.loader = loader
        self.path_to_flatten_metadata = (
            "model_artifacts/system_store/flatten_configs/"
            f"flatten_metadata_{fetch_unique_root(self.table_name, self.metadata_path)}.json"
        )


class PreprocessHandler(Processor):
    """
    The class for the preprocessing of the data before the training process
    """
    def __init__(
        self,
        metadata: Dict,
        metadata_path: Optional[str],
        table_name: str,
        loader: Optional[Callable[[str], pd.DataFrame]] = None
    ):
        super().__init__(
            metadata=metadata,
            metadata_path=metadata_path,
            table_name=table_name,
            loader=loader
        )
        self.initial_data_shape: Tuple = ()
        self.row_subset: int = int()
        self.original_df: Optional[pd.DataFrame] = None
        self.initial_order_of_columns = list()
        self.schema = dict()
        self.original_schema = None

    def _save_original_schema(self):
        """
        Save the schema of the original data
        """
        path = (
            f"model_artifacts/tmp_store/{slugify(self.table_name)}"
            f"/original_schema_{slugify(self.table_name)}.pkl"
        )
        DataLoader(path).save_data(data=self.original_schema)

    def _save_initial_order_of_columns(self):
        """
        Save the initial order of columns of the original data
        """
        path = (
            f"model_artifacts/tmp_store/{slugify(self.table_name)}"
            f"/initial_order_of_columns_{slugify(self.table_name)}.pkl"
        )
        DataLoader(path).save_data(data=self.initial_order_of_columns)

    def _check_if_data_is_empty(self):
        """
        Check if the provided data is empty or not
        """
        if self.original_df.shape[0] < 1:
            raise ValueError(
                f"The empty table was provided. Unable to train the table - '{self.table_name}'"
            )

    def _remove_empty_columns(self) -> pd.DataFrame:
        """
        Remove completely empty columns from dataframe and mark them as 'removed' in the schema
        """
        data = self.original_df.dropna(how="all", axis=1)

        dropped_columns = set(self.initial_order_of_columns) - set(data.columns)
        list_of_dropped_columns = [f"'{column}'" for column in dropped_columns]
        if list_of_dropped_columns:
            logger.info(f"Empty columns - {', '.join(list_of_dropped_columns)} were removed")
            self.schema["fields"].update({column: "removed" for column in dropped_columns})
        return data

    def prepare_data(self) -> pd.DataFrame:
        """
        Prepare the subset of the data for the training process,
        and get the preprocessed data
        """
        self._load_source()
        self._check_if_data_is_empty()
        self._save_original_schema()
        self._save_initial_order_of_columns()
        preprocessed_data = self._preprocess_data()
        return preprocessed_data

    def _preprocess_data(self) -> pd.DataFrame:
        """
        Apply the parameters 'drop_null' and 'row_limit' to the data,
        and get the subset for the training process
        """
        drop_null = self.metadata[self.table_name]["train_settings"]["drop_null"]
        row_limit = self.metadata[self.table_name]["train_settings"]["row_limit"]
        data = self._remove_empty_columns()
        if self.loader:
            warning_message = (
                "parameter will be ignored because the retrieval of the data "
                "is handled by a callback function"
            )
            if drop_null:
                logger.warning(f"The 'drop_null' {warning_message}")
            if row_limit is not None:
                logger.warning(f"The 'row_limit' {warning_message}")
        else:
            if drop_null:
                if not data.dropna().empty:
                    data = data.dropna()
                    if count_of_dropped_rows := self.initial_data_shape[0] - data.shape[0]:
                        logger.info(
                            f"As the parameter 'drop_null' set to 'True', "
                            f"{count_of_dropped_rows} rows of the table - '{self.table_name}' "
                            f"that have empty values have been dropped. "
                            f"The count of remained rows is {data.shape[0]}."
                        )
                else:
                    logger.warning(
                        "The specified 'drop_null' argument results in the empty dataframe, "
                        "so it will be ignored"
                    )
                self.row_subset = len(data)

            if row_limit:
                self.row_subset = min(row_limit, len(data))
                data = data.sample(n=self.row_subset)

        if len(data) < 100:
            logger.warning(
                "The input table is too small to provide any meaningful results. "
                "Please consider: 1) disable drop_null argument, 2) provide bigger table"
            )
        elif len(data) < 500:
            logger.warning(
                f"The amount of data is {len(data)} rows. It seems that it isn't enough "
                f"to supply high-quality results. To improve the quality of generated data "
                f"please consider any of the steps: 1) provide a bigger table, "
                f"2) disable drop_null argument"
            )
        logger.info(f"The subset of rows was set to {self.row_subset}")
        return data

    @staticmethod
    def _run_script():
        """
        Run the script before the training process
        if it exists in the predefined path
        """
        path_to_script = f"{os.getcwd()}/model_artifacts/script.py"
        if os.path.exists(path_to_script):
            os.system(f"python3 {path_to_script}")

    @staticmethod
    def _get_json_columns(data: pd.DataFrame) -> List[str]:
        """
        Get the list of columns which contain JSON data convertable to a Python dictionary
        Returns:
            List[str]: List of column names containing valid JSON data convertable to dict
        """
        def is_json_column(series: pd.Series) -> bool:
            if pd.isnull(series).all():
                return False

            def is_json_value(x):
                if not isinstance(x, str):
                    return False
                try:
                    return isinstance(json.loads(x), dict)
                except JSONDecodeError:
                    return False

            return series.dropna().apply(is_json_value).any()
        return [col for col in data.columns if is_json_column(data[col])]

    @staticmethod
    def _get_artifacts(
        data: pd.DataFrame,
        json_columns: List[str]
    ) -> Tuple[pd.DataFrame, Dict[str, List[str]], List[str]]:
        """
        Flatten JSON columns in DataFrame and create mapping of original to flattened columns.

        Args:
            data (pd.DataFrame): Input DataFrame with JSON columns
            json_columns (List[str]): List of columns containing JSON data

        Returns:
            Tuple containing:
            - Flattened DataFrame
            - Mapping of original columns to flattened columns
            - List of duplicated column names
        """
        flattened_dfs = list()
        flattening_mapping = dict()
        for column in json_columns:
            result = [safe_flatten(i) for i in data[column]]
            flattened_data = pd.DataFrame(
                [i["flattened_data"] for i in result],
                index=data.index
            )
            flattened_data[f"{column}_"] = pd.DataFrame(
                [i["original_data"] for i in result],
                index=data.index
            )
            flattening_mapping[column] = [
                c for c in flattened_data.columns.to_list()
                if c != f"{column}_"
            ]
            flattened_dfs.append(flattened_data)
        flattened_data = pd.concat([data, *flattened_dfs], axis=1)
        duplicated_columns = [
            key
            for key, value in dict(Counter(flattened_data.columns.to_list())).items()
            if value > 1
        ]
        flattened_data.drop(columns=flattening_mapping.keys(), inplace=True)
        flattened_df = flattened_data.T.loc[~flattened_data.T.index.duplicated(), :].T
        flattened_df = flattened_df.applymap(lambda x: np.NaN if x in [list(), dict()] else x)
        return flattened_df, flattening_mapping, duplicated_columns

    def _save_flatten_metadata(self, metadata: Dict):
        """
        Save the metadata of the flattening process
        """
        if os.path.exists(self.path_to_flatten_metadata):
            with open(self.path_to_flatten_metadata, "r") as f:
                existing_metadata = json.load(f)
            metadata = {**existing_metadata, **metadata}
        with open(f"{self.path_to_flatten_metadata}", "w") as f:
            json.dump(metadata, f)

    def _load_source(self):
        """
        Load the data from the predefined source
        """
        if self.loader is not None:
            data_loader = DataFrameFetcher(
                loader=self.loader,
                table_name=self.table_name
            )
        else:
            path_to_source = self.metadata[self.table_name]["train_settings"]["source"]
            data_loader = DataLoader(path=path_to_source)
        self.original_df, self.schema = data_loader.load_data()
        self.original_schema = data_loader.original_schema
        self.initial_data_shape = self.original_df.shape
        self.initial_order_of_columns = self.original_df.columns.to_list()
        self.row_subset = len(self.original_df)

    def _handle_json_columns(self, data: pd.DataFrame):
        """
        Preprocess the data contained JSON columns before the training process
        """
        flatten_metadata = dict()
        order_of_columns = data.columns.to_list()
        if json_columns := self._get_json_columns(data):
            list_of_json_columns = [f"'{column}'" for column in json_columns]
            list_of_json_columns = ', '.join(list_of_json_columns)
            logger.info(
                f"The table '{self.table_name}' contains JSON columns: {list_of_json_columns}"
            )
            logger.info(f"Flattening the JSON columns in the table - '{self.table_name}'")
            (
                flattened_data,
                flattening_mapping,
                duplicated_columns
            ) = self._get_artifacts(data, json_columns)
            flatten_metadata[self.table_name] = {
                "flattening_mapping": flattening_mapping,
                "duplicated_columns": duplicated_columns,
                "order_of_columns": order_of_columns
            }
            logger.info(f"The table '{self.table_name}' has been successfully flattened")
            self._save_flatten_metadata(flatten_metadata)
            return flattened_data
        return data

    def run(self) -> Tuple[pd.DataFrame, Dict]:
        """
        Launch the preprocessing process:
        """
        self._run_script()
        preprocessed_data = self.prepare_data()
        data = self._handle_json_columns(preprocessed_data)
        return data, self.schema


class PostprocessHandler(Processor):
    """
    Class for the postprocessing of the generated data
    """

    def __init__(
        self,
        metadata: Dict,
        metadata_path: Optional[str],
        table_name: Optional[str],
        type_of_process: Literal["train", "infer"]
    ):
        super().__init__(metadata, metadata_path, table_name)
        self.type_of_process = type_of_process
        self.flatten_metadata = self._fetch_flatten_config()
        self.preview_rows = 10

    def _fetch_flatten_config(self) -> Dict:
        """
        Fetch the metadata of the flattening process
        """
        if os.path.exists(self.path_to_flatten_metadata):
            with open(self.path_to_flatten_metadata, "r") as f:
                return json.load(f)
        return dict()

    @staticmethod
    def _check_none_values(x: Any) -> bool:
        """
        Check if the value should be treated as null/none.

        Args:
            x: Value to check

        Returns:
            bool: True if value should be treated as null/none

        Note:
            Handles None, np.NaN, and empty containers as null values
        """
        if x is None:
            return True
        if isinstance(x, float) and np.isnan(x):
            return True
        if isinstance(x, (list, dict)) and not x:
            return True
        return False

    def _remove_none_from_struct(self, input_data: Any) -> Any:
        """
        Remove np.NaN or None values from nested data structures.

        Args:
            input_data: Input data structure (list, dict, or scalar)

        Returns:
            Cleaned data structure with None values removed
        """
        if isinstance(input_data, list):
            # Process list elements and filter empty results
            return [
                self._remove_none_from_struct(v)
                for v in input_data
                if not self._check_none_values(v)
            ]

        if isinstance(input_data, dict):
            # Process dictionary items and filter empty results
            return {
                k: self._remove_none_from_struct(v)
                for k, v in input_data.items()
                if not self._check_none_values(v)
            }
        return input_data

    @staticmethod
    def _restore_empty_values(df: pd.DataFrame) -> pd.DataFrame:
        """
        Restore empty dictionary values in the dataframe for nested fields.

        Args:
            df: Input DataFrame with potentially empty nested fields

        Returns:
            pd.DataFrame: DataFrame with restored empty dictionary values
        """
        df = df.copy().astype(object)
        columns = df.columns.tolist()

        # Find potential parent columns
        parent_columns = set()
        for col in columns:
            for potential_parent in columns:
                if f"{potential_parent}." in col:
                    parent_columns.add(potential_parent)

        # Restore empty dictionary values for parent columns
        for parent_col in parent_columns:
            df[parent_col] = df[parent_col].apply(lambda val: {} if pd.isna(val) else val)

        return df

    @staticmethod
    def _load_generated_data(path_to_generated_data: str, *args) -> pd.DataFrame:
        """
        Load generated data from the predefined path
        """
        data, schema = DataLoader(path=path_to_generated_data).load_data()
        return data

    @staticmethod
    def _remove_empty_elements(d: Union[Dict, List]) -> Optional[Union[Dict, List]]:
        """
        Recursively remove keys with empty dictionaries or lists from a nested dictionary or list
        """
        def clean(data):
            if isinstance(data, dict):
                cleaned = {
                    key: cleaned_value
                    for key, value in data.items()
                    if (cleaned_value := clean(value)) not in ({}, [])
                }
                return cleaned

            if isinstance(data, list):
                cleaned = [
                    cleaned_item
                    for item in data
                    if (cleaned_item := clean(item)) not in ({}, [])
                ]
                return cleaned

            return data
        result = clean(d)
        return result if pd.notnull(result) and result not in [{}, []] else None

    def _postprocess_generated_data(
        self,
        data: pd.DataFrame,
        flattening_mapping: Dict,
        duplicated_columns: List
    ) -> pd.DataFrame:
        """
        Postprocess the generated data
        """
        for old_column, new_columns in flattening_mapping.items():
            data[new_columns] = self._restore_empty_values(data[new_columns])

            data[old_column] = data.apply(
                lambda row: unflatten_list(row[new_columns].to_dict(), ".")
                if row[new_columns].notnull().any()
                else row[f"{old_column}_"],
                axis=1
            )
            data[old_column] = data[old_column].apply(
                lambda row: self._remove_none_from_struct(row)
            )
            data[old_column] = data[old_column].apply(lambda row: self._remove_empty_elements(row))
            data[old_column] = data[old_column].apply(
                lambda row: json.dumps(row, ensure_ascii=False)
                if isinstance(row, dict)
                else row
            )
            dropped_columns = set(i for i in new_columns if i not in duplicated_columns)
            dropped_columns.add(f"{old_column}_")
            data.drop(list(dropped_columns), axis=1, inplace=True)
        return data

    def run(self):
        """
        Launch the postprocessing of the generated data,
        and save the processed data to the predefined path
        """
        for table in self.metadata.keys():
            path_to_merged_infer = fetch_config(
                f"model_artifacts/resources/{slugify(table)}/vae/checkpoints/train_config.pkl"
            ).paths["path_to_merged_infer"]
            extension = get_source_path_extension(path=path_to_merged_infer)
            default_path_to_generated_data = (
                f"model_artifacts/tmp_store/{slugify(table)}/"
                f"merged_infer_{slugify(table)}{extension}"
            )
            path_to_generated_data = (
                self.metadata[table].get("infer_settings", {}).get(
                    "destination", default_path_to_generated_data
                ) if self.type_of_process == "infer" else default_path_to_generated_data
            )
            data = self._load_generated_data(path_to_generated_data, table)
            flatten_metadata_of_table = self.flatten_metadata.get(table)
            if flatten_metadata_of_table is not None:
                logger.info("Start postprocessing of the generated data")
                flattening_mapping = flatten_metadata_of_table.get("flattening_mapping")
                duplicated_columns = flatten_metadata_of_table.get("duplicated_columns")
                data = self._postprocess_generated_data(
                    data,
                    flattening_mapping,
                    duplicated_columns
                )
                logger.info("Finish postprocessing of the generated data")
            order_of_columns = fetch_config(
                config_pickle_path=f"model_artifacts/tmp_store/{slugify(table)}"
                                   f"/initial_order_of_columns_{slugify(table)}.pkl")
            self.__save_generated_data(
                data,
                path_to_generated_data,
                order_of_columns,
                table
            )

    @staticmethod
    def __save_preview_data(
        path_to_destination,
        preview_df
    ):
        """
        Save the preview data in '.csv' format for the fast review of the generated data
        """
        preview_path = f"{os.path.splitext(path_to_destination)[0]}_preview.csv"
        DataLoader(path=preview_path).save_data(
            preview_df,
            format=get_context().get_config(),
        )
        logger.info(
            f"Preview saved in '{preview_path}' (first {len(preview_df)} rows)"
        )

    def __save_generated_data(
        self,
        generated_data: pd.DataFrame,
        path_to_destination: str,
        order_of_columns: List[str],
        *args
    ):
        """
        Save generated data to the path
        """
        generated_data = generated_data[order_of_columns]
        DataLoader(path=path_to_destination).save_data(
            generated_data,
            format=get_context().get_config(),
        )
        if self.type_of_process == "infer":
            self.__save_preview_data(
                path_to_destination=path_to_destination,
                preview_df=generated_data.head(self.preview_rows)
            )
